"""
This module contains models for kudosu objects.
"""
from __future__ import annotations

from datetime import datetime
from typing import Literal
from typing import Optional

from .base import BaseModel

__all__ = (
    "KudosuAction",
    "KudosuGiver",
    "KudosuPost",
    "KudosuHistory",
)

KudosuAction = Literal[
    "give",
    "vote.give",
    "reset",
    "vote.reset",
    "revoke",
    "vote.revoke",
]


class KudosuGiver(BaseModel):
    url: str
    username: str


class KudosuPost(BaseModel):
    title: str
    url: Optional[str]


class KudosuHistory(BaseModel):
    id: int
    action: KudosuAction
    created_at: datetime
    amount: int
    model: str
    giver: Optional[KudosuGiver]
    post: Optional[KudosuPost]



"""This module contains a repository for API v2 token objects."""
from __future__ import annotations

from abc import ABC
from abc import abstractmethod

from ...models import OAuthToken

__all__ = ("BaseTokenRepository", "SimpleTokenRepository")


class BaseTokenRepository(ABC):
    """Base token repository. Allows for custom token storage."""

    @abstractmethod
    async def exists(self, session_id: int) -> bool:
        """Check if a token exists."""
        ...

    @abstractmethod
    async def add(self, session_id: int, token: OAuthToken) -> OAuthToken:
        """Add a token."""
        ...

    @abstractmethod
    async def get(self, session_id: int) -> OAuthToken:
        """Get a token."""
        ...

    @abstractmethod
    async def update(self, session_id: int, token: OAuthToken) -> OAuthToken:
        """Update a token."""
        ...

    @abstractmethod
    async def delete(self, session_id: int) -> None:
        """Delete a token."""
        ...


class SimpleTokenRepository(BaseTokenRepository):
    """Simple in-memory token repository."""

    __slots__ = ("_tokens",)

    def __init__(self) -> None:
        self._tokens: dict[int, OAuthToken] = {}

    async def exists(self, session_id: int) -> bool:
        """Check if a token exists."""
        return session_id in self._tokens

    async def add(self, session_id: int, token: OAuthToken) -> OAuthToken:
        """Add a token."""
        self._tokens[session_id] = token
        return token

    async def get(self, session_id: int) -> OAuthToken:
        """Get a token."""
        return self._tokens[session_id]

    async def update(self, session_id: int, token: OAuthToken) -> OAuthToken:
        """Update a token."""
        self._tokens[session_id] = token
        return token

    async def delete(self, session_id: int) -> None:
        """Delete a token."""
        del self._tokens[session_id]



"""
This module contains models for API v2 difficulty attribute objects.
"""
from __future__ import annotations

import abc

from .base import BaseModel

__all__ = (
    "PerformanceAttributes",
    "OsuPerformanceAttributes",
    "TaikoPerformanceAttributes",
    "ManiaPerformanceAttributes",
    "CatchPerformanceAttributes",
)


class PerformanceAttributes(BaseModel, abc.ABC):
    total: float


class OsuPerformanceAttributes(PerformanceAttributes):
    aim: float
    speed: float
    accuracy: float
    flashlight: float
    effective_miss_count: float


class TaikoPerformanceAttributes(PerformanceAttributes):
    difficulty: float
    accuracy: float
    effective_miss_count: float


class ManiaPerformanceAttributes(PerformanceAttributes):
    difficulty: float


class CatchPerformanceAttributes(PerformanceAttributes):
    ...



"""
This module contains models for Beatmap objects.
"""
from __future__ import annotations

from datetime import datetime
from enum import Enum
from enum import unique
from typing import Any
from typing import Literal
from typing import Optional

from pydantic import Field
from pydantic import root_validator

from .base import BaseModel
from .common import CursorModel
from .gamemode import Gamemode
from .user import User

__all__ = (
    "Beatmap",
    "BeatmapAvailability",
    "BeatmapCovers",
    "BeatmapDifficultyAttributes",
    "BeatmapFailtimes",
    "BeatmapHype",
    "BeatmapNominations",
    "BeatmapRankStatus",
    "Beatmapset",
    "BeatmapsetDiscussion",
    "BeatmapsetDiscussionPost",
    "BeatmapsetDisscussionType",
    "BeatmapsetEvent",
    "BeatmapsetEventComment",
    "BeatmapsetEventType",
    "BeatmapsetRequestStatus",
    "BeatmapsetVoteEvent",
    "BeatmapUserPlaycount",
    "BeatmapsetDiscussionResponse",
    "BeatmapsetDiscussionPostResponse",
    "BeatmapsetDiscussionVoteResponse",
    "BeatmapsetSearchResponse",
)

BeatmapsetDisscussionType = Literal[
    "hype",
    "praise",
    "problem",
    "review",
    "suggestion",
    "mapper_note",
]


BeatmapsetEventType = Literal[
    "approve",
    "beatmap_owner_change",
    "discussion_delete",
    "discussion_post_delete",
    "discussion_post_restore",
    "discussion_restore",
    "discussion_lock",
    "discussion_unlock",
    "disqualify",
    "genre_edit",
    "issue_reopen",
    "issue_resolve",
    "kudosu_allow",
    "kudosu_deny",
    "kudosu_gain",
    "kudosu_lost",
    "kudosu_recalculate",
    "language_edit",
    "love",
    "nominate",
    "nomination_reset",
    "nomination_reset_received",
    "nsfw_toggle",
    "offset_edit",
    "qualify",
    "rank",
    "remove_from_loved",
]

BeatmapsetRequestStatus = Literal[
    "all",
    "ranked",
    "qualified",
    "disqualified",
    "never_ranked",
]

BEATMAP_RANK_STATUS_NAMES = {
    -2: "graveyard",
    -1: "wip",
    0: "pending",
    1: "ranked",
    2: "approved",
    3: "qualified",
    4: "loved",
}


@unique
class BeatmapRankStatus(Enum):
    GRAVEYARD = -2
    WIP = -1
    PENDING = 0
    RANKED = 1
    APPROVED = 2
    QUALIFIED = 3
    LOVED = 4

    @property
    def id(self) -> int:
        return self.value

    @property
    def name_api(self) -> str:
        return BEATMAP_RANK_STATUS_NAMES[self.id]

    def __str__(self) -> str:
        return self.name_api

    @classmethod
    def _missing_(cls, query: object) -> Any:
        if isinstance(query, int):
            for status in list(BeatmapRankStatus):
                if status.id == query:
                    return status
        elif isinstance(query, str):
            for status in list(BeatmapRankStatus):
                if status.name_api == query.lower():
                    return status
        raise ValueError(f"BeatmapRankStatus {query} does not exist.")


class BeatmapAvailability(BaseModel):
    more_information: Optional[str]
    download_disabled: Optional[bool]

    @classmethod
    def _from_api_v1(cls, data: Any) -> BeatmapAvailability:
        return cls.parse_obj({"download_disabled": data["download_unavailable"]})


class BeatmapNominations(BaseModel):
    current: Optional[int]
    required: Optional[int]


class BeatmapCovers(BaseModel):
    cover: str
    card: str
    list: str
    slimcover: str
    cover_2_x: Optional[str] = Field(alias="cover@2x")
    card_2_x: Optional[str] = Field(alias="card@2x")
    list_2_x: Optional[str] = Field(alias="list@2x")
    slimcover_2_x: Optional[str] = Field(alias="slimcover@2x")

    @classmethod
    def from_beatmapset_id(cls, beatmapset_id: int) -> BeatmapCovers:
        base_url = "https://assets.ppy.sh/beatmaps/"
        return cls.parse_obj(
            {
                "cover": f"{base_url}{beatmapset_id}/covers/cover.jpg",
                "card": f"{base_url}{beatmapset_id}/covers/card.jpg",
                "list": f"{base_url}{beatmapset_id}/covers/list.jpg",
                "slimcover": f"{base_url}{beatmapset_id}/covers/slimcover.jpg",
                "cover_2_x": f"{base_url}{beatmapset_id}/covers/cover@2x.jpg",
                "card_2_x": f"{base_url}{beatmapset_id}/covers/card@2x.jpg",
                "list_2_x": f"{base_url}{beatmapset_id}/covers/list@2x.jpg",
                "slimcover_2_x": f"{base_url}{beatmapset_id}/covers/slimcover@2x.jpg",
            },
        )

    @classmethod
    def _from_api_v1(cls, data: Any) -> BeatmapCovers:
        return cls.from_beatmapset_id(data["beatmapset_id"])


class BeatmapHype(BaseModel):
    current: int
    required: int


class BeatmapFailtimes(BaseModel):
    exit: Optional[list[int]]
    fail: Optional[list[int]]


class BeatmapDifficultyAttributes(BaseModel):
    max_combo: int
    star_rating: float
    # osu standard
    aim_difficulty: Optional[float]
    approach_rate: Optional[float]  # osu catch + standard
    flashlight_difficulty: Optional[float]
    overall_difficulty: Optional[float]
    slider_factor: Optional[float]
    speed_difficulty: Optional[float]
    speed_note_count: Optional[float]
    # osu taiko
    stamina_difficulty: Optional[float]
    rhythm_difficulty: Optional[float]
    colour_difficulty: Optional[float]
    # osu mania
    great_hit_window: Optional[float]
    score_multiplier: Optional[float]


class Beatmap(BaseModel):
    id: int
    url: str
    mode: Gamemode
    beatmapset_id: int
    difficulty_rating: float
    status: BeatmapRankStatus
    total_length: int
    user_id: int
    version: str
    accuracy: Optional[float]
    ar: Optional[float]
    cs: Optional[float]
    bpm: Optional[float]
    convert: Optional[bool]
    count_circles: Optional[int]
    count_sliders: Optional[int]
    count_spinners: Optional[int]
    deleted_at: Optional[datetime]
    drain: Optional[float]
    hit_length: Optional[int]
    is_scoreable: Optional[bool]
    last_updated: Optional[datetime]
    passcount: Optional[int]
    play_count: Optional[int] = Field(None, alias="playcount")
    checksum: Optional[str]
    max_combo: Optional[int]
    beatmapset: Optional[Beatmapset]
    failtimes: Optional[BeatmapFailtimes]

    @root_validator(pre=True)
    def _set_url(cls, values: dict[str, Any]) -> dict[str, Any]:
        if values.get("url") is None:
            id = values["id"]
            beatmapset_id = values["beatmapset_id"]
            mode = Gamemode(values["mode"])
            values[
                "url"
            ] = f"https://osu.ppy.sh/beatmapsets/{beatmapset_id}#{mode}/{id}"
        return values

    @property
    def discussion_url(self) -> str:
        return f"https://osu.ppy.sh/beatmapsets/{self.beatmapset_id}/discussion/{self.id}/general"

    @property
    def count_objects(self) -> int:
        """Total count of the objects.

        :raises ValueError: Raised if object counts are none
        :return: Sum of counts of all objects
        :rtype: int
        """
        if (
            self.count_circles is None
            or self.count_spinners is None
            or self.count_sliders is None
        ):
            raise ValueError("Beatmap contains no object count information.")
        return self.count_spinners + self.count_circles + self.count_sliders

    @classmethod
    def _from_api_v1(cls, data: Any) -> Beatmap:
        return cls.parse_obj(
            {
                "beatmapset_id": data["beatmapset_id"],
                "difficulty_rating": data["difficultyrating"],
                "id": data["beatmap_id"],
                "mode": int(data["mode"]),
                "status": int(data["approved"]),
                "total_length": data["total_length"],
                "hit_length": data["total_length"],
                "user_id": data["creator_id"],
                "version": data["version"],
                "accuracy": data["diff_overall"],
                "cs": data["diff_size"],
                "ar": data["diff_approach"],
                "drain": data["diff_drain"],
                "last_updated": data["last_update"],
                "bpm": data["bpm"],
                "checksum": data["file_md5"],
                "playcount": data["playcount"],
                "passcount": data["passcount"],
                "count_circles": data["count_normal"],
                "count_sliders": data["count_slider"],
                "count_spinners": data["count_spinner"],
                "max_combo": data["max_combo"],
            },
        )


class Beatmapset(BaseModel):
    id: int
    artist: str
    artist_unicode: str
    covers: BeatmapCovers
    creator: str
    favourite_count: int
    play_count: int = Field(alias="playcount")
    preview_url: str
    source: str
    status: BeatmapRankStatus
    title: str
    title_unicode: str
    user_id: int
    video: bool
    nsfw: Optional[bool]
    hype: Optional[BeatmapHype]
    availability: Optional[BeatmapAvailability]
    bpm: Optional[float]
    can_be_hyped: Optional[bool]
    discussion_enabled: Optional[bool]
    discussion_locked: Optional[bool]
    is_scoreable: Optional[bool]
    last_updated: Optional[datetime]
    legacy_thread_url: Optional[str]
    nominations_summary: Optional[BeatmapNominations]
    ranked_date: Optional[datetime]
    storyboard: Optional[bool]
    submitted_date: Optional[datetime]
    tags: Optional[str]
    ratings: Optional[list[int]]
    has_favourited: Optional[bool]
    beatmaps: Optional[list[Beatmap]]

    @property
    def url(self) -> str:
        return f"https://osu.ppy.sh/beatmapsets/{self.id}"

    @property
    def discussion_url(self) -> str:
        return f"https://osu.ppy.sh/beatmapsets/{self.id}/discussion"

    @classmethod
    def _from_api_v1(cls, data: Any) -> Beatmapset:
        return cls.parse_obj(
            {
                "id": data["beatmapset_id"],
                "artist": data["artist"],
                "artist_unicode": data["artist"],
                "covers": BeatmapCovers._from_api_v1(data),
                "favourite_count": data["favourite_count"],
                "creator": data["creator"],
                "play_count": data["playcount"],
                "preview_url": f"//b.ppy.sh/preview/{data['beatmapset_id']}.mp3",
                "source": data["source"],
                "status": int(data["approved"]),
                "title": data["title"],
                "title_unicode": data["title"],
                "user_id": data["creator_id"],
                "video": data["video"],
                "submitted_date": data["submit_date"],
                "ranked_date": data["approved_date"],
                "last_updated": data["last_update"],
                "tags": data["tags"],
                "storyboard": data["storyboard"],
                "availabiliy": BeatmapAvailability._from_api_v1(data),
                "beatmaps": [Beatmap._from_api_v1(data)],
            },
        )


class BeatmapsetSearchResponse(CursorModel):
    beatmapsets: list[Beatmapset]


class BeatmapUserPlaycount(BaseModel):
    count: int
    beatmap_id: int
    beatmap: Optional[Beatmap]
    beatmapset: Optional[Beatmapset]


class BeatmapsetDiscussionPost(BaseModel):
    id: int
    user_id: int
    system: bool
    message: str
    created_at: datetime
    beatmap_discussion_id: Optional[int]
    last_editor_id: Optional[int]
    deleted_by_id: Optional[int]
    updated_at: Optional[datetime]
    deleted_at: Optional[datetime]


class BeatmapsetDiscussion(BaseModel):
    id: int
    beatmapset_id: int
    user_id: int
    message_type: BeatmapsetDisscussionType
    resolved: bool
    can_be_resolved: bool
    can_grant_kudosu: bool
    created_at: datetime
    beatmap_id: Optional[int]
    deleted_by_id: Optional[int]
    parent_id: Optional[int]
    timestamp: Optional[int]
    updated_at: Optional[datetime]
    deleted_at: Optional[datetime]
    last_post_at: Optional[datetime]
    kudosu_denied: Optional[bool]
    starting_post: Optional[BeatmapsetDiscussionPost]


class BeatmapsetVoteEvent(BaseModel):
    score: int
    user_id: int
    id: Optional[int]
    created_at: Optional[datetime]
    updated_at: Optional[datetime]
    beatmapset_discussion_id: Optional[int]


class BeatmapsetEventComment(BaseModel):
    beatmap_discussion_id: Optional[int]
    beatmap_discussion_post_id: Optional[int]
    new_vote: Optional[BeatmapsetVoteEvent]
    votes: Optional[list[BeatmapsetVoteEvent]]
    mode: Optional[Gamemode]
    reason: Optional[str]
    source_user_id: Optional[int]
    source_user_username: Optional[str]
    nominator_ids: Optional[list[int]]
    new: Optional[str]
    old: Optional[str]
    new_user_id: Optional[int]
    new_user_username: Optional[str]


class BeatmapsetEvent(BaseModel):
    id: int
    type: BeatmapsetEventType
    r"""Information on types: https://github.com/ppy/osu-web/blob/master/resources/assets/lib/interfaces/beatmapset-event-json.ts"""
    created_at: datetime
    user_id: int
    beatmapset: Optional[Beatmapset]
    discussion: Optional[BeatmapsetDiscussion]
    comment: Optional[dict]


class BeatmapsetDiscussionResponse(CursorModel):
    beatmaps: list[Beatmap]
    discussions: list[BeatmapsetDiscussion]
    included_discussions: list[BeatmapsetDiscussion]
    users: list[User]
    max_blocks: int

    @root_validator(pre=True)
    def _set_max_blocks(cls, values: dict[str, Any]) -> dict[str, Any]:
        values["max_blocks"] = values["reviews_config"]["max_blocks"]
        return values


class BeatmapsetDiscussionPostResponse(CursorModel):
    beatmapsets: list[Beatmapset]
    posts: list[BeatmapsetDiscussionPost]
    users: list[User]


class BeatmapsetDiscussionVoteResponse(CursorModel):
    votes: list[BeatmapsetVoteEvent]
    discussions: list[BeatmapsetDiscussion]
    users: list[User]


Beatmap.update_forward_refs()



from __future__ import annotations

from ..base import BaseModel

__all__ = ("ReplayCompact",)


class ReplayCompact(BaseModel):
    """Replay API object."""

    content: str  # test
    """Encoded LZMA data of the replay"""
    encoding: str
    """The encoding used"""



"""
This module contains models for Gamemode objects.
"""
from __future__ import annotations

from enum import Enum
from enum import unique
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any

__all__ = ("Gamemode",)

GAMEMODE_NAMES = {
    0: "Standard",
    1: "Taiko",
    2: "Catch the Beat",
    3: "Mania",
}

GAMEMODE_SHORT_NAMES = {
    0: "STD",
    1: "Taiko",
    2: "CTB",
    3: "Mania",
}

GAMEMODE_API_NAMES = {
    0: "osu",
    1: "taiko",
    2: "fruits",
    3: "mania",
}


@unique
class Gamemode(Enum):
    STANDARD = 0
    TAIKO = 1
    CTB = 2
    MANIA = 3

    @property
    def id(self) -> int:
        return self.value

    @property
    def name_full(self) -> str:
        return GAMEMODE_NAMES[self.id]

    @property
    def name_short(self) -> str:
        return GAMEMODE_SHORT_NAMES[self.id]

    @property
    def name_api(self) -> str:
        return GAMEMODE_API_NAMES[self.id]

    def __int__(self) -> int:
        return self.id

    def __str__(self) -> str:
        return self.name_api

    def __format__(self, spec: str) -> str:
        if spec == "f":
            return self.name_full
        if spec == "s":
            return self.name_short
        return self.name_api

    @classmethod
    def from_type(cls, __o: object) -> Gamemode:
        """Gets a gamemode.

        :param __o: Object to search for
        :type __o: object
        :raises ValueError: If object cannot be converted to Gamemode
        :return: A Gamemode object. Will search by name_api, name_short, id
        :rtype: aiosu.models.gamemode.Gamemode
        """
        if isinstance(__o, cls):
            return __o
        for mode in list(Gamemode):
            if __o in (mode.name_api, mode.name_short, mode.name_full, mode.id):
                return mode
        raise ValueError(f"Gamemode {__o} does not exist.")

    @classmethod
    def _missing_(cls, query: object) -> Any:
        return cls.from_type(query)



"""
This module contains models for User objects.
"""
from __future__ import annotations

from datetime import datetime
from enum import Enum
from enum import unique
from typing import Any
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from pydantic import Field

from .base import BaseModel
from .common import Country
from .common import HTMLBody
from .common import TimestampedCount
from .gamemode import Gamemode

if TYPE_CHECKING:
    from typing import Callable

__all__ = (
    "User",
    "UserAccountHistory",
    "UserBadge",
    "UserGradeCounts",
    "UserGroup",
    "UserKudosu",
    "UserLevel",
    "UserProfileCover",
    "UserProfileTournamentBanner",
    "UserQueryType",
    "UserRankHistoryElement",
    "UserStats",
    "UserAccountHistoryType",
    "UserRankHighest",
)


cast_int: Callable[..., int] = lambda x: int(x or 0)
cast_float: Callable[..., float] = lambda x: float(x or 0)

UserAccountHistoryType = Literal[
    "note",
    "restriction",
    "silence",
    "tournament_ban",
]

OLD_QUERY_TYPES = {
    "ID": "id",
    "USERNAME": "string",
}


@unique
class UserQueryType(Enum):
    ID = "id"
    USERNAME = "username"

    @property
    def old_api_name(self) -> str:
        return OLD_QUERY_TYPES[self.name]

    @property
    def new_api_name(self) -> str:
        return self.value

    @classmethod
    def _missing_(cls, query: object) -> Any:
        if isinstance(query, str):
            query = query.lower()
        for q in list(UserQueryType):
            if query in (q.old_api_name, q.new_api_name):
                return q
        raise ValueError(f"UserQueryType {query} does not exist.")


class UserLevel(BaseModel):
    current: int
    progress: float

    @classmethod
    def _from_api_v1(cls, data: Any) -> UserLevel:
        level = cast_float(data["level"])
        current = int(level)
        progress = (level - current) * 100
        return cls.parse_obj({"current": current, "progress": progress})


class UserKudosu(BaseModel):
    total: int
    available: int


class UserRankHistoryElement(BaseModel):
    mode: str
    data: list[int]

    @property
    def average_gain(self) -> float:
        r"""Average rank gain.

        :return: Average rank gain for a user
        :rtype: float
        """
        return (self.data[1] - self.data[-1]) / len(self.data)


class UserRankHighest(BaseModel):
    rank: int
    updated_at: datetime


class UserProfileCover(BaseModel):
    url: str
    custom_url: Optional[str]
    id: Optional[str]


class UserProfileTournamentBanner(BaseModel):
    tournament_id: int
    id: Optional[int]
    image: Optional[str]
    image_2_x: Optional[str] = Field(alias="image@2x")


class UserBadge(BaseModel):
    awarded_at: datetime
    description: str
    image_url: str
    url: str


class UserAccountHistory(BaseModel):
    id: int
    timestamp: datetime
    length: int
    permanent: bool
    type: UserAccountHistoryType
    description: Optional[str]


class UserGradeCounts(BaseModel):
    ssh: Optional[int]
    """Number of Silver SS ranks achieved."""
    ss: Optional[int]
    """Number of SS ranks achieved."""
    sh: Optional[int]
    """Number of Silver S ranks achieved."""
    s: Optional[int]
    """Number of S ranks achieved."""
    a: Optional[int]
    """Number of A ranks achieved."""

    @classmethod
    def _from_api_v1(cls, data: Any) -> UserGradeCounts:
        return cls.parse_obj(
            {
                "ss": cast_int(data["count_rank_ss"]),
                "ssh": cast_int(data["count_rank_ssh"]),
                "s": cast_int(data["count_rank_s"]),
                "sh": cast_int(data["count_rank_sh"]),
                "a": cast_int(data["count_rank_a"]),
            },
        )


class UserGroup(BaseModel):
    id: int
    identifier: str
    name: str
    short_name: str
    has_listing: bool
    has_playmodes: bool
    is_probationary: bool
    colour: Optional[str]
    playmodes: Optional[list[Gamemode]]
    description: Optional[str]


class UserStats(BaseModel):
    """Fields are marked as optional since they might be missing from rankings other than performance."""

    ranked_score: Optional[int]
    play_count: Optional[int]
    grade_counts: Optional[UserGradeCounts]
    total_hits: Optional[int]
    is_ranked: Optional[bool]
    total_score: Optional[int]
    level: Optional[UserLevel]
    hit_accuracy: Optional[float]
    play_time: Optional[int]
    pp: Optional[float]
    pp_exp: Optional[float]
    replays_watched_by_others: Optional[int]
    maximum_combo: Optional[int]
    global_rank: Optional[int]
    global_rank_exp: Optional[int]
    country_rank: Optional[int]
    user: Optional[User]
    count_300: Optional[int]
    count_100: Optional[int]
    count_50: Optional[int]
    count_miss: Optional[int]

    @property
    def pp_per_playtime(self) -> float:
        r"""PP per playtime.

        :return: PP per playtime
        :rtype: float
        """
        if not self.play_time or not self.pp:
            return 0
        return self.pp / self.play_time * 3600

    @classmethod
    def _from_api_v1(cls, data: Any) -> UserStats:
        """Some fields can be None, we want to force them to cast to a value."""
        return cls.parse_obj(
            {
                "level": UserLevel._from_api_v1(data),
                "pp": cast_float(data["pp_raw"]),
                "global_rank": cast_int(data["pp_rank"]),
                "country_rank": cast_int(data["pp_country_rank"]),
                "ranked_score": cast_int(data["ranked_score"]),
                "hit_accuracy": cast_float(data["accuracy"]),
                "play_count": cast_int(data["playcount"]),
                "play_time": cast_int(data["total_seconds_played"]),
                "total_score": cast_int(data["total_score"]),
                "total_hits": cast_int(data["count300"])
                + cast_int(data["count100"])
                + cast_int(data["count50"]),
                "is_ranked": cast_float(data["pp_raw"]) != 0,
                "grade_counts": UserGradeCounts._from_api_v1(data),
                "count_300": cast_int(data["count300"]),
                "count_100": cast_int(data["count100"]),
                "count_50": cast_int(data["count50"]),
            },
        )


class UserAchievmement(BaseModel):
    achieved_at: datetime
    achievement_id: int


class User(BaseModel):
    avatar_url: str
    country_code: str
    id: int
    username: str
    default_group: Optional[str]
    is_active: Optional[bool]
    is_bot: Optional[bool]
    is_online: Optional[bool]
    is_supporter: Optional[bool]
    pm_friends_only: Optional[bool]
    profile_colour: Optional[str]
    is_deleted: Optional[bool]
    last_visit: Optional[datetime]
    discord: Optional[str]
    has_supported: Optional[bool]
    interests: Optional[str]
    join_date: Optional[datetime]
    kudosu: Optional[UserKudosu]
    location: Optional[str]
    max_blocks: Optional[int]
    max_friends: Optional[int]
    occupation: Optional[str]
    playmode: Optional[Gamemode]
    playstyle: Optional[list[str]]
    post_count: Optional[int]
    profile_order: Optional[list[str]]
    title: Optional[str]
    twitter: Optional[str]
    website: Optional[str]
    country: Optional[Country]
    cover: Optional[UserProfileCover]
    is_restricted: Optional[bool]
    account_history: Optional[list[UserAccountHistory]]
    active_tournament_banner: Optional[UserProfileTournamentBanner]
    badges: Optional[list[UserBadge]]
    beatmap_playcounts_count: Optional[int]
    favourite_beatmapset_count: Optional[int]
    follower_count: Optional[int]
    graveyard_beatmapset_count: Optional[int]
    groups: Optional[list[UserGroup]]
    loved_beatmapset_count: Optional[int]
    monthly_playcounts: Optional[list[TimestampedCount]]
    page: Optional[HTMLBody]
    pending_beatmapset_count: Optional[int]
    previous_usernames: Optional[list[str]]
    ranked_beatmapset_count: Optional[int]
    replays_watched_counts: Optional[list[TimestampedCount]]
    scores_best_count: Optional[int]
    scores_first_count: Optional[int]
    scores_recent_count: Optional[int]
    statistics: Optional[UserStats]
    support_level: Optional[int]
    user_achievements: Optional[list[UserAchievmement]]
    rank_history: Optional[UserRankHistoryElement]
    rank_highest: Optional[UserRankHighest]

    @property
    def url(self) -> str:
        return f"https://osu.ppy.sh/users/{self.id}"

    @classmethod
    def _from_api_v1(cls, data: Any) -> User:
        return cls.parse_obj(
            {
                "avatar_url": f"https://s.ppy.sh/a/{data['user_id']}",
                "country_code": data["country"],
                "id": data["user_id"],
                "username": data["username"],
                "join_date": data["join_date"],
                "statistics": UserStats._from_api_v1(data),
            },
        )


UserStats.update_forward_refs()



"""
This module contains models for API v2 token objects.
"""
from __future__ import annotations

from datetime import datetime
from datetime import timedelta
from functools import cached_property
from typing import TYPE_CHECKING

import jwt
from pydantic import root_validator

from .base import FrozenModel
from .scopes import Scopes

if TYPE_CHECKING:
    from typing import Any

__all__ = ("OAuthToken",)


class OAuthToken(FrozenModel, keep_untouched=(cached_property,)):
    token_type: str = "Bearer"
    """Defaults to 'Bearer'"""
    access_token: str = ""
    refresh_token: str = ""
    expires_on: datetime = datetime.utcfromtimestamp(0)
    """Can be a datetime.datetime object or a string. Alternatively, expires_in may be passed representing the number of seconds the token will be valid for."""

    @cached_property
    def owner_id(self) -> int:
        if not self.access_token:
            return 0
        decoded = jwt.decode(self.access_token, options={"verify_signature": False})
        if decoded["sub"]:
            return int(decoded["sub"])
        return 0

    @cached_property
    def scopes(self) -> Scopes:
        if not self.access_token:
            return Scopes.PUBLIC
        decoded = jwt.decode(self.access_token, options={"verify_signature": False})
        return Scopes.from_api_list(decoded["scopes"])

    @cached_property
    def can_refresh(self) -> bool:
        """Returns True if the token can be refreshed."""
        return bool(self.refresh_token)

    @root_validator(pre=True)
    def _set_expires_on(cls, values: dict[str, Any]) -> dict[str, Any]:
        if isinstance(values.get("expires_in"), int):
            values["expires_on"] = datetime.utcnow() + timedelta(
                seconds=values["expires_in"],
            )
        return values



"""
This module contains base classes for library events.
"""
from __future__ import annotations

import abc
import dataclasses
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Type
    from typing import Callable
    from .models import OAuthToken

__all__ = (
    "Eventable",
    "BaseEvent",
    "ClientAddEvent",
    "ClientUpdateEvent",
)


class Eventable(abc.ABC):
    """Abstract for classes that handle events"""

    __slots__ = ("_listeners",)

    def __init__(self) -> None:
        self._listeners: dict[str, list[Callable]] = {}

    def _register_event(self, event: Type[BaseEvent]) -> None:
        r"""Registers an event

        :param event: Event type to register
        :type event: Type[BaseEvent]
        """
        self._listeners[event._name] = []

    def _register_listener(self, func: Callable, event: Type[BaseEvent]) -> None:
        r"""Registers an event listener

        :param func: Function to call when event is emitted
        :type func: Callable
        :param event: Event type to listen for
        :type event: Type[BaseEvent]

        :raises NotImplementedError: If event is not implemented
        """
        if event._name not in self._listeners:
            raise NotImplementedError(f"{event!r}")
        self._listeners[event._name].append(func)

    async def _process_event(self, event: BaseEvent) -> None:
        r"""Processes an event

        :param event: Event to process
        :type event: BaseEvent

        :raises NotImplementedError: If event is not implemented
        """
        if event._name not in self._listeners:
            raise NotImplementedError(f"{event!r}")
        for listener in self._listeners[event._name]:
            await listener(event)


@dataclasses.dataclass
class BaseEvent(abc.ABC):
    """Abstract for event classes"""

    _name = "BaseEvent"


@dataclasses.dataclass
class ClientAddEvent(BaseEvent):
    """Event for when a client is added"""

    _name = "ClientAddEvent"
    session_id: int
    """0 if app client"""
    client: Eventable
    """The client that was added"""


@dataclasses.dataclass
class ClientUpdateEvent(BaseEvent):
    """Event for when a client is updated"""

    _name = "ClientUpdateEvent"
    client: Eventable
    old_token: OAuthToken
    new_token: OAuthToken



"""
This module contains authorization functions.
"""
from __future__ import annotations

import urllib.parse
from typing import Optional

import aiohttp
import orjson

from ..exceptions import APIException
from ..models.oauthtoken import OAuthToken
from ..models.scopes import Scopes

__all__ = (
    "generate_url",
    "process_code",
)


async def process_code(
    client_id: int,
    client_secret: str,
    redirect_uri: str,
    code: str,
    base_url: str = "https://osu.ppy.sh",
) -> OAuthToken:
    r"""Creates an OAuth Token from an authorization code.

    :param client_id: The ID of the client
    :type client_id: int
    :param client_secret: The client secret
    :type client_secret: str
    :param redirect_uri: The URL to redirect to
    :type redirect_uri: str
    :param code: Code returned from the API
    :type code: str
    :param base_url: The base URL of the API, defaults to "https://osu.ppy.sh"
    :type base_url: Optional[str]
    :return: The OAuth token
    :rtype: aiosu.models.oauthtoken.OAuthToken
    """
    url = f"{base_url}/oauth/token"
    headers = {"Content-Type": "application/x-www-form-urlencoded"}
    data = {
        "client_id": client_id,
        "client_secret": client_secret,
        "code": code,
        "redirect_uri": redirect_uri,
        "grant_type": "authorization_code",
    }

    async with aiohttp.ClientSession(headers=headers) as temp_session:
        async with temp_session.post(url, data=data) as resp:
            try:
                body = await resp.read()
                json = orjson.loads(body)
                if resp.status != 200:
                    raise APIException(resp.status, json.get("error", ""))
                token = OAuthToken.parse_obj(json)
                return token
            except aiohttp.client_exceptions.ContentTypeError:
                raise APIException(403, "Invalid code specified.")


def generate_url(
    client_id: int,
    redirect_uri: str,
    base_url: str = "https://osu.ppy.sh",
    scopes: Scopes = Scopes.PUBLIC | Scopes.IDENTIFY,
    state: Optional[str] = None,
) -> str:
    r"""Generates an OAuth URL.

    :param client_id: The ID of the client
    :type client_id: int
    :param redirect_uri: The URL to redirect to
    :type redirect_uri: str
    :param base_url: The base URL of the API, defaults to "https://osu.ppy.sh"
    :type base_url: Optional[str]
    :param scopes: The scopes to request, defaults to Scopes.PUBLIC | Scopes.IDENTIFY
    :type scopes: Optional[Scopes]
    :param state: The state to pass to the API, defaults to None
    :type state: Optional[str]
    :return: The OAuth URL
    :rtype: str
    """
    params = {
        "client_id": client_id,
        "redirect_uri": redirect_uri,
        "response_type": "code",
        "scope": str(scopes),
    }
    if state:
        params["state"] = state
    return f"{base_url}/oauth/authorize?{urllib.parse.urlencode(params)}"



"""
This module contains performance point calculators for osu! gamemodes.
"""
from __future__ import annotations

import abc
import math
from typing import Callable
from typing import TYPE_CHECKING

from ..models import CatchPerformanceAttributes
from ..models import Gamemode
from ..models import ManiaPerformanceAttributes
from ..models import Mod
from ..models import OsuPerformanceAttributes
from ..models import TaikoPerformanceAttributes
from .accuracy import CatchAccuracyCalculator
from .accuracy import ManiaAccuracyCalculator
from .accuracy import OsuAccuracyCalculator
from .accuracy import TaikoAccuracyCalculator

if TYPE_CHECKING:
    from typing import Type
    from ..models.score import Score
    from ..models.beatmap import BeatmapDifficultyAttributes
    from ..models.performance import PerformanceAttributes


__all__ = [
    "OsuPerformanceCalculator",
    "TaikoPerformanceCalculator",
    "ManiaPerformanceCalculator",
    "CatchPerformanceCalculator",
]

OSU_BASE_MULTIPLIER = 1.14
TAIKO_BASE_MULTIPLIER = 1.13
MANIA_BASE_MULTIPLIER = 8.0
CATCH_BASE_MULTIPLIER = 1.0

clamp: Callable[[float, float, float], float] = (
    lambda x, l, u: l if x < l else u if x > u else x
)


class AbstractPerformanceCalculator(abc.ABC):
    __slots__ = ("difficulty_attributes",)

    def __init__(self, difficulty_attributes: BeatmapDifficultyAttributes):
        self.difficulty_attributes = difficulty_attributes

    @abc.abstractmethod
    def calculate(self, score: Score) -> PerformanceAttributes:
        ...


def get_calculator(mode: Gamemode) -> Type[AbstractPerformanceCalculator]:
    r"""Returns the performance calculator for the given gamemode.

    :param mode: The gamemode to get the calculator for
    :type mode: aiosu.models.gamemode.Gamemode
    :raises ValueError: If the gamemode is unknown
    :return: The performance calculator type for the given gamemode
    :rtype: Type[AbstractPerformanceCalculator]
    """
    if mode == Gamemode.STANDARD:
        return OsuPerformanceCalculator
    elif mode == Gamemode.TAIKO:
        return TaikoPerformanceCalculator
    elif mode == Gamemode.MANIA:
        return ManiaPerformanceCalculator
    elif mode == Gamemode.CTB:
        return CatchPerformanceCalculator
    else:
        raise ValueError(f"Unknown gamemode: {mode}")


class OsuPerformanceCalculator(AbstractPerformanceCalculator):
    r"""osu!std performance point calculator. Only compatible with scores from API v2.

    :param difficulty_attributes: API difficulty attributes for a beatmap
    :type difficulty_attributes: BeatmapDifficultyAttributes
    """

    def calculate(self, score: Score) -> OsuPerformanceAttributes:
        r"""Calculates performance points for a score.

        :param score: The score to calculate pp for
        :type score: aiosu.models.score.Score
        :raises ValueError: If score does not have an associated beatmap
        :return: Performance attributes for the score
        :rtype: aiosu.models.performance.OsuPerformanceAttributes
        """
        if score.beatmap is None:
            raise ValueError("Given score does not have a beatmap.")

        effective_miss_count = self._calculate_effective_miss_count(score)
        total_hits = (
            score.statistics.count_300
            + score.statistics.count_100
            + score.statistics.count_50
            + score.statistics.count_miss
        )

        multiplier = OSU_BASE_MULTIPLIER

        if Mod.NoFail in score.mods:
            multiplier *= max(0.9, 1.0 - 0.02 * effective_miss_count)

        if Mod.SpunOut in score.mods and total_hits > 0:
            multiplier *= 1.0 - math.pow(
                (score.beatmap.count_spinners / total_hits),  # type: ignore
                0.85,
            )

        aim_value = self._compute_aim_value(score, effective_miss_count, total_hits)
        speed_value = self._compute_speed_value(score, effective_miss_count, total_hits)
        accuracy_value = self._compute_accuracy_value(score, total_hits)
        flashlight_value = self._compute_flashlight_value(
            score,
            effective_miss_count,
            total_hits,
        )

        total_value = (
            math.pow(
                math.pow(aim_value, 1.1)
                + math.pow(speed_value, 1.1)
                + math.pow(accuracy_value, 1.1)
                + math.pow(flashlight_value, 1.1),
                1.0 / 1.1,
            )
            * multiplier
        )

        return OsuPerformanceAttributes(
            total=total_value,
            aim=aim_value,
            speed=speed_value,
            accuracy=accuracy_value,
            flashlight=flashlight_value,
            effective_miss_count=effective_miss_count,
        )

    def _compute_aim_value(
        self,
        score: Score,
        effective_miss_count: float,
        total_hits: int,
    ) -> float:
        aim_value = (
            math.pow(
                5.0 * max(1.0, self.difficulty_attributes.aim_difficulty / 0.0675)  # type: ignore
                - 4.0,
                3.0,
            )
            / 100000.0
        )

        length_bonus = (
            0.95
            + 0.4 * min(1.0, total_hits / 2000.0)
            + ((math.log10(total_hits / 2000.0) * 0.5) * int(total_hits > 2000))
        )
        aim_value *= length_bonus

        if effective_miss_count > 0:
            aim_value *= 0.97 * math.pow(
                1 - math.pow(effective_miss_count / total_hits, 0.775),
                effective_miss_count,
            )

        aim_value *= self._get_combo_scaling_factor(score)

        approach_rate_factor = 0.0
        if self.difficulty_attributes.approach_rate > 10.33:  # type: ignore
            approach_rate_factor = 0.3 * (
                self.difficulty_attributes.approach_rate - 10.33  # type: ignore
            )
        elif self.difficulty_attributes.approach_rate < 8.0:  # type: ignore
            approach_rate_factor = 0.05 * (
                8.0 - self.difficulty_attributes.approach_rate  # type: ignore
            )

        aim_value *= 1.0 + approach_rate_factor * length_bonus

        if Mod.Hidden in score.mods:
            aim_value *= 1.0 + 0.04 * (12.0 - self.difficulty_attributes.approach_rate)  # type: ignore

        if score.beatmap.count_sliders > 0:  # type: ignore
            estimate_difficult_sliders = score.beatmap.count_sliders * 0.15  # type: ignore

            estimate_slider_ends_dropped = clamp(
                min(
                    score.statistics.count_100
                    + score.statistics.count_50
                    + score.statistics.count_miss,
                    self.difficulty_attributes.max_combo - score.max_combo,
                ),
                0,
                estimate_difficult_sliders,
            )

            slider_nerf_factor = (  # type: ignore
                1 - self.difficulty_attributes.slider_factor  # type: ignore
            ) * math.pow(
                1 - estimate_slider_ends_dropped / estimate_difficult_sliders,
                3,
            ) + self.difficulty_attributes.slider_factor

            aim_value *= slider_nerf_factor

        accuracy = score.accuracy if score.accuracy <= 1.0 else score.accuracy / 100
        aim_value *= accuracy
        aim_value *= (
            0.98 + math.pow(self.difficulty_attributes.overall_difficulty, 2) / 2500  # type: ignore
        )

        return aim_value

    def _compute_speed_value(
        self,
        score: Score,
        effective_miss_count: float,
        total_hits: int,
    ) -> float:
        speed_value = (
            math.pow(
                5.0 * max(1.0, self.difficulty_attributes.speed_difficulty / 0.0675)  # type: ignore
                - 4.0,
                3.0,
            )
            / 100000.0
        )

        length_bonus = (
            0.95
            + 0.4 * min(1.0, total_hits / 2000.0)
            + ((math.log10(total_hits / 2000.0) * 0.5) * int(total_hits > 2000))
        )
        speed_value *= length_bonus

        if effective_miss_count > 0:
            speed_value *= 0.97 * math.pow(
                1 - math.pow(effective_miss_count / total_hits, 0.775),
                math.pow(effective_miss_count, 0.875),
            )

        speed_value *= self._get_combo_scaling_factor(score)

        approach_rate_factor = 0.0
        if self.difficulty_attributes.approach_rate > 10.33:  # type: ignore
            approach_rate_factor = 0.3 * (
                self.difficulty_attributes.approach_rate - 10.33  # type: ignore
            )

        speed_value *= 1.0 + approach_rate_factor * length_bonus

        if Mod.Hidden in score.mods:
            speed_value *= 1.0 + 0.04 * (
                12.0 - self.difficulty_attributes.approach_rate  # type: ignore
            )

        relevant_total_diff = total_hits - self.difficulty_attributes.speed_note_count  # type: ignore
        relevant_count_great = max(0, score.statistics.count_300 - relevant_total_diff)
        relevant_count_ok = max(
            0,
            score.statistics.count_100
            - max(0, relevant_total_diff - score.statistics.count_300),
        )
        relevant_count_meh = max(
            0,
            score.statistics.count_50
            - max(
                0,
                relevant_total_diff
                - score.statistics.count_300
                - score.statistics.count_100,
            ),
        )

        relevant_accuracy = 0
        if self.difficulty_attributes.speed_note_count > 0:  # type: ignore
            relevant_accuracy = (
                relevant_count_great * 6.0
                + relevant_count_ok * 2.0
                + relevant_count_meh
            ) / (
                self.difficulty_attributes.speed_note_count * 6.0  # type: ignore
            )

        accuracy = score.accuracy if score.accuracy <= 1.0 else score.accuracy / 100

        speed_value *= (
            0.95 + math.pow(self.difficulty_attributes.overall_difficulty, 2) / 750  # type: ignore
        ) * math.pow(
            (accuracy + relevant_accuracy) / 2.0,
            (14.5 - max(self.difficulty_attributes.overall_difficulty, 8)) / 2,  # type: ignore
        )

        speed_value *= math.pow(
            0.99,
            (score.statistics.count_50 - total_hits / 500.0)
            * int(score.statistics.count_50 > total_hits / 500.0),
        )

        return speed_value

    def _compute_accuracy_value(
        self,
        score: Score,
        total_hits: int,
    ) -> float:
        accuracy_calculator = OsuAccuracyCalculator()
        better_accuracy_percentage = accuracy_calculator.calculate_weighted(score)

        accuracy_value = (
            math.pow(1.52163, self.difficulty_attributes.overall_difficulty)  # type: ignore
            * math.pow(better_accuracy_percentage, 24)
            * 2.83
        )

        accuracy_value *= min(
            1.15,
            math.pow(score.beatmap.count_circles / 1000.0, 0.3),  # type: ignore
        )

        if Mod.Hidden in score.mods:
            accuracy_value *= 1.08

        if Mod.Flashlight in score.mods:
            accuracy_value *= 1.02

        return accuracy_value

    def _compute_flashlight_value(
        self,
        score: Score,
        effective_miss_count: float,
        total_hits: int,
    ) -> float:
        if Mod.Flashlight not in score.mods:
            return 0.0

        flashlight_value = (
            math.pow(self.difficulty_attributes.flashlight_difficulty, 2.0) * 25.0  # type: ignore
        )

        if effective_miss_count > 0:
            flashlight_value *= 0.97 * math.pow(
                1 - math.pow(effective_miss_count / total_hits, 0.775),
                math.pow(effective_miss_count, 0.875),
            )

        flashlight_value *= self._get_combo_scaling_factor(score)

        flashlight_value *= (
            0.7
            + 0.1 * min(1.0, total_hits / 200.0)
            + 0.2 * (min(1.0, (total_hits - 200) / 200.0) * int(total_hits > 200))
        )

        accuracy = score.accuracy if score.accuracy <= 1.0 else score.accuracy / 100
        flashlight_value *= 0.5 + accuracy / 2.0
        flashlight_value *= (
            0.98 + math.pow(self.difficulty_attributes.overall_difficulty, 2) / 2500.0  # type: ignore
        )

        return flashlight_value

    def _calculate_effective_miss_count(self, score: Score) -> float:
        combo_based_miss_count = 0.0

        if score.beatmap.count_sliders > 0:  # type: ignore
            full_combo_threshold = (
                self.difficulty_attributes.max_combo - 0.1 * score.beatmap.count_sliders  # type: ignore
            )

            if score.max_combo < full_combo_threshold:
                combo_based_miss_count = full_combo_threshold / max(
                    1.0,
                    score.max_combo,
                )

        combo_based_miss_count = min(
            combo_based_miss_count,
            score.statistics.count_100
            + score.statistics.count_50
            + score.statistics.count_miss,
        )

        return max(score.statistics.count_miss, combo_based_miss_count)

    def _get_combo_scaling_factor(self, score: Score) -> float:
        if self.difficulty_attributes.max_combo <= 0:
            return 1.0

        return min(
            math.pow(score.max_combo, 0.8)
            / math.pow(self.difficulty_attributes.max_combo, 0.8),
            1.0,
        )


class TaikoPerformanceCalculator(AbstractPerformanceCalculator):
    r"""osu!taiko performance point calculator.

    :param difficulty_attributes: API difficulty attributes for a beatmap
    :type difficulty_attributes: BeatmapDifficultyAttributes
    """

    def calculate(self, score: Score) -> TaikoPerformanceAttributes:
        r"""Calculates performance points for a score

        :param score: The score to calculate pp for
        :type score: aiosu.models.score.Score
        :return: Performance attributes for the score
        :rtype: aiosu.models.performance.TaikoPerformanceAttributes
        """
        accuracy_calculator = TaikoAccuracyCalculator()
        accuracy = accuracy_calculator.calculate_weighted(score)

        effective_miss_count = self._calculate_effective_miss_count(score)
        total_hits = (
            score.statistics.count_300
            + score.statistics.count_100
            + score.statistics.count_50
            + score.statistics.count_miss
        )

        multiplier = TAIKO_BASE_MULTIPLIER

        if Mod.Hidden in score.mods:
            multiplier *= 1.075

        if Mod.Easy in score.mods:
            multiplier *= 0.975

        difficulty_value = self._compute_difficulty_value(
            score,
            total_hits,
            effective_miss_count,
            accuracy,
        )
        accuracy_value = self._compute_accuracy_value(
            score,
            total_hits,
            accuracy,
        )
        total_value = (
            math.pow(
                math.pow(difficulty_value, 1.1) + math.pow(accuracy_value, 1.1),
                1.0 / 1.1,
            )
            * multiplier
        )

        return TaikoPerformanceAttributes(
            total=total_value,
            difficulty=difficulty_value,
            accuracy=accuracy_value,
            effective_miss_count=effective_miss_count,
        )

    def _compute_difficulty_value(
        self,
        score: Score,
        total_hits: int,
        effective_miss_count: float,
        accuracy: float,
    ) -> float:
        difficulty_value = (
            math.pow(
                5 * max(1.0, self.difficulty_attributes.star_rating / 0.115) - 4.0,
                2.25,
            )
            / 1150.0
        )

        length_bonus = 1 + 0.1 * min(1.0, total_hits / 1500.0)
        difficulty_value *= length_bonus

        difficulty_value *= math.pow(0.986, effective_miss_count)

        if Mod.Easy in score.mods:
            difficulty_value *= 0.985

        if Mod.Hidden in score.mods:
            difficulty_value *= 1.025

        if Mod.HardRock in score.mods:
            difficulty_value *= 1.050

        if Mod.Flashlight in score.mods:
            difficulty_value *= 1.050 * length_bonus

        difficulty_value *= math.pow(accuracy, 2.0)
        return difficulty_value

    def _compute_accuracy_value(
        self,
        score: Score,
        total_hits: int,
        accuracy: float,
    ) -> float:
        if self.difficulty_attributes.great_hit_window <= 0:  # type: ignore
            return 0

        accuracy_value = (
            math.pow(60.0 / self.difficulty_attributes.great_hit_window, 1.1)  # type: ignore
            * math.pow(accuracy, 8.0)
            * math.pow(self.difficulty_attributes.star_rating, 0.4)
            * 27.0
        )

        length_bonus = min(1.15, math.pow(total_hits / 1500.0, 0.3))
        accuracy_value *= length_bonus

        if Mod.Hidden in score.mods and Mod.Flashlight in score.mods:
            accuracy_value *= max(1.050, 1.075 * length_bonus)

        return accuracy_value

    def _calculate_effective_miss_count(self, score: Score) -> float:
        return (
            max(
                1.0,
                1000.0
                / (
                    score.statistics.count_300
                    + score.statistics.count_100
                    + score.statistics.count_50
                ),
            )
            * score.statistics.count_miss
        )


class ManiaPerformanceCalculator(AbstractPerformanceCalculator):
    r"""osu!mania performance point calculator.

    :param difficulty_attributes: API difficulty attributes for a beatmap
    :type difficulty_attributes: BeatmapDifficultyAttributes
    """

    def calculate(self, score: Score) -> ManiaPerformanceAttributes:
        r"""Calculates performance points for a score.

        :param score: The score to calculate pp for
        :type score: aiosu.models.score.Score
        :return: Performance attributes for the score
        :rtype: aiosu.models.performance.ManiaPerformanceAttributes
        """
        accuracy_calculator = ManiaAccuracyCalculator()
        accuracy = accuracy_calculator.calculate_weighted(score)

        total_hits = (
            score.statistics.count_geki
            + score.statistics.count_300
            + score.statistics.count_katu
            + score.statistics.count_100
            + score.statistics.count_50
            + score.statistics.count_miss
        )

        multiplier = MANIA_BASE_MULTIPLIER

        if Mod.NoFail in score.mods:
            multiplier *= 0.75

        if Mod.Easy in score.mods:
            multiplier *= 0.5

        difficulty_value = self._compute_difficulty_value(accuracy, total_hits)
        total_value = difficulty_value * multiplier

        return ManiaPerformanceAttributes(
            total=total_value,
            difficulty=difficulty_value,
        )

    def _compute_difficulty_value(self, accuracy: float, total_hits: int) -> float:
        difficulty_value = (
            math.pow(max(self.difficulty_attributes.star_rating - 0.15, 0.05), 2.2)
            * max(0.0, 5.0 * accuracy - 4.0)
            * (1.0 + 0.1 * min(1.0, total_hits / 1500))
        )

        return difficulty_value


class CatchPerformanceCalculator(AbstractPerformanceCalculator):
    r"""osu!catch performance point calculator.

    :param difficulty_attributes: API difficulty attributes for a beatmap
    :type difficulty_attributes: BeatmapDifficultyAttributes
    """

    def calculate(self, score: Score) -> CatchPerformanceAttributes:
        r"""Calculates performance points for a score.

        :param score: The score to calculate pp for
        :type score: aiosu.models.score.Score
        :return: Performance attributes for the score
        :rtype: aiosu.models.performance.CatchPerformanceAttributes
        """
        accuracy_calculator = CatchAccuracyCalculator()
        accuracy = accuracy_calculator.calculate_weighted(score)

        total_combo_hits = (
            score.statistics.count_miss
            + score.statistics.count_100
            + score.statistics.count_300
        )

        multiplier = CATCH_BASE_MULTIPLIER

        total_value = (
            math.pow(
                5.0 * max(1.0, self.difficulty_attributes.star_rating / 0.0049) - 4.0,
                2.0,
            )
            / 100000.0
        )

        length_bonus = (
            0.95
            + 0.3 * min(1.0, total_combo_hits / 2500.0)
            + (
                (math.log10(total_combo_hits / 2500.0) * 0.475)
                * int(total_combo_hits > 2500)
            )
        )
        total_value *= length_bonus

        total_value *= math.pow(0.97, score.statistics.count_miss)

        if self.difficulty_attributes.max_combo > 0:
            total_value *= min(
                math.pow(score.max_combo, 0.8)
                / math.pow(self.difficulty_attributes.max_combo, 0.8),
                1.0,
            )

        approach_rate: float = self.difficulty_attributes.approach_rate  # type: ignore
        approach_rate_factor = 1.0

        if approach_rate > 9.0:
            approach_rate_factor += 0.1 * (approach_rate - 9.0)

        if approach_rate > 10.0:
            approach_rate_factor += 0.1 * (approach_rate - 10.0)
        elif approach_rate < 8.0:
            approach_rate_factor += 0.025 * (8.0 - approach_rate)

        total_value *= approach_rate_factor

        if Mod.Hidden in score.mods:
            if approach_rate <= 10.0:
                total_value *= 1.05 + 0.075 * (10.0 - approach_rate)
            elif approach_rate > 10.0:
                total_value *= 1.01 + 0.04 * (11.0 - min(11.0, approach_rate))

        if Mod.Flashlight in score.mods:
            total_value *= 1.35 * length_bonus

        total_value *= math.pow(accuracy, 5.5)

        if Mod.NoFail in score.mods:
            total_value *= 0.90

        total_value *= multiplier

        return CatchPerformanceAttributes(total=total_value)



"""
This module contains models for chat.
"""
from __future__ import annotations

from typing import Literal
from typing import Optional

from pydantic import Field

from .base import BaseModel
from .common import CurrentUserAttributes
from .user import User

__all__ = (
    "ChatChannel",
    "ChatMessage",
    "ChatMessageCreateResponse",
    "ChatChannelTypes",
    "ChatIncludeTypes",
    "ChatUpdateResponse",
    "ChatUserSilence",
    "ChatChannelResponse",
)

ChatChannelTypes = Literal[
    "PM",
    "PUBLIC",
    "PRIVATE",
    "MULTIPLAYER",
    "SPECTATOR",
    "TEMPORARY",
    "GROUP",
    "ANNOUNCE",
]

ChatIncludeTypes = Literal[
    "messages",
    "presence",
    "silences",
]


class ChatUserSilence(BaseModel):
    id: int
    user_id: int


class ChatChannel(BaseModel):
    id: int = Field(alias="channel_id")
    type: ChatChannelTypes
    name: str
    moderated: bool
    icon: Optional[str]
    description: Optional[str]
    last_message_id: Optional[int]
    user_ids: Optional[list[int]] = Field(alias="users")
    current_user_attributes: Optional[CurrentUserAttributes]


class ChatMessage(BaseModel):
    message_id: int
    sender_id: int
    channel_id: int
    timestamp: str
    content: str
    is_action: bool
    uuid: Optional[str]
    sender: Optional[User]


class ChatMessageCreateResponse(BaseModel):
    channel: ChatChannel
    message: ChatMessage


class ChatUpdateResponse(BaseModel):
    messages: Optional[list[ChatMessage]]
    presence: Optional[list[ChatChannel]]
    silences: list


class ChatChannelResponse(BaseModel):
    channel: ChatChannel
    users: Optional[list[User]]



"""
This module contains base models for objects.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import orjson
import pydantic

from .mods import Mods


if TYPE_CHECKING:
    from typing import Any

__all__ = (
    "BaseModel",
    "FrozenModel",
)


def orjson_dumps(v: object, *, default: Any) -> str:
    # orjson.dumps returns bytes, to match standard json.dumps we need to decode
    return orjson.dumps(v, default=default).decode()


class BaseModel(pydantic.BaseModel):
    class Config:
        arbitrary_types_allowed = True
        allow_population_by_field_name = True
        json_loads = orjson.loads
        json_dumps = orjson_dumps

        json_encoders = {
            Mods: str,
        }


class FrozenModel(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self.__config__.frozen = True



"""
This module contains models for multiplayer.
"""
from __future__ import annotations

from datetime import datetime
from typing import Any
from typing import Literal
from typing import Optional

from pydantic import Field
from pydantic import root_validator

from .base import BaseModel
from .beatmap import Beatmap
from .common import CursorModel
from .gamemode import Gamemode
from .lazer import LazerMod
from .lazer import LazerScoreStatistics
from .user import User

__all__ = (
    "MultiplayerScoreSortType",
    "MultiplayerScoresResponse",
    "MultiplayerScore",
    "MultiplayerScoresAround",
    "MultiplayerMatch",
    "MultiplayerEventTypes",
    "MultiplayerEvent",
    "MultiplayerMatchResponse",
    "MultiplayerMatchesResponse",
    "MultiplayerRoomMode",
    "MultiplayerRoom",
    "MultiplayerRoomsResponse",
    "MultiplayerRoomCategories",
    "MultiplayerRoomTypeGroups",
    "MultiplayerLeaderboardResponse",
    "MultiplayerLeaderboardItem",
    "MultiplayerQueueMode",
)

MultiplayerScoreSortType = Literal["score_asc", "score_desc"]
MultiplayerEventTypes = Literal[
    "match-created",
    "match-disbanded",
    "host-changed",
    "player-joined",
    "player-left",
    "player-kicked",
    "match-created-no-user",
    "match-disbanded-no-user",
    "host-changed-no-user",
    "player-joined-no-user",
    "player-left-no-user",
    "player-kicked-no-user",
]
MultiplayerRoomMode = Literal["owned", "participated", "ended"]
MultiplayerRoomCategories = Literal["normal", "spotlight", "featured_artists"]
MultiplayerRoomTypeGroups = Literal["playlists", "realtime"]
MultiplayerQueueMode = Literal["host_only", "all_players", "all_players_round_robin"]


class MultiplayerScoresAround(BaseModel):
    higher: MultiplayerScoresResponse
    lower: MultiplayerScoresResponse


class MultiplayerScore(BaseModel):
    id: int
    user_id: int
    room_id: int
    playlist_item_id: int
    beatmap_id: int
    rank: str
    total_score: int
    accuracy: float
    max_combo: int
    mods: list[LazerMod]
    passed: bool
    user: User
    statistics: LazerScoreStatistics
    position: Optional[int]
    scores_around: Optional[MultiplayerScoresAround]


class MultiplayerScoresResponse(CursorModel):
    scores: list[MultiplayerScore]
    user_score: Optional[MultiplayerScore]
    total: Optional[int]


class MultiplayerMatch(BaseModel):
    id: int
    name: str
    start_time: datetime
    end_time: Optional[datetime]


class MultiplayerEvent(BaseModel):
    id: int
    timestamp: datetime
    type: MultiplayerEventTypes
    user_id: Optional[int]

    @root_validator(pre=True)
    def _set_type(cls, values: dict[str, Any]) -> dict[str, Any]:
        if "detail" in values:
            values["type"] = values["detail"]["type"]
        return values


class MultiplayerMatchResponse(BaseModel):
    match: MultiplayerMatch
    events: list[MultiplayerEvent]
    users: list[User]
    first_event_id: int
    latest_event_id: int
    current_game_id: Optional[int]


class MultiplayerMatchesResponse(CursorModel):
    matches: list[MultiplayerMatch]


class MultiplayerPlaylistItem(BaseModel):
    id: int
    room_id: int
    beatmap_id: int
    mode: Gamemode = Field(alias="ruleset_id")
    allowed_mods: list[LazerMod]
    required_mods: list[LazerMod]
    expired: bool
    owner_id: int
    beatmap: Beatmap
    playlist_order: Optional[int]
    played_at: Optional[datetime]


class MultiplayerRoom(BaseModel):
    id: int
    name: str
    category: MultiplayerRoomCategories
    type: MultiplayerRoomTypeGroups
    user_id: int
    channel_id: int
    active: bool
    has_password: bool
    auto_skip: bool
    host: User
    queue_mode: MultiplayerQueueMode
    playlist: list[MultiplayerPlaylistItem]
    recent_participants: list[User]
    participant_count: Optional[int]
    starts_at: Optional[datetime]
    ends_at: Optional[datetime]
    max_attempts: Optional[int]


class MultiplayerLeaderboardItem(BaseModel):
    accuracy: float
    attempts: int
    completed: int
    pp: float
    room_id: int
    total_score: int
    user_id: int
    user: User
    position: Optional[int]


class MultiplayerLeaderboardResponse(BaseModel):
    leaderboard: list[MultiplayerLeaderboardItem]
    user_score: Optional[MultiplayerLeaderboardItem]


class MultiplayerRoomsResponse(CursorModel):
    """Currently unused. Relevant for api-version >= 99999999"""

    rooms: list[MultiplayerRoom]


MultiplayerScoresAround.update_forward_refs()



"""
This module contains models for miscellaneous objects.
"""
from __future__ import annotations

from datetime import datetime
from functools import partial
from typing import Any
from typing import Coroutine
from typing import Literal
from typing import Optional

from emojiflags.lookup import lookup as flag_lookup  # type: ignore
from pydantic import Field
from pydantic import validator

from .base import BaseModel
from .gamemode import Gamemode

__all__ = (
    "Achievement",
    "Country",
    "CurrentUserAttributes",
    "TimestampedCount",
    "CursorModel",
    "SortTypes",
    "HTMLBody",
)


SortTypes = Literal["id_asc", "id_desc"]


class TimestampedCount(BaseModel):
    start_date: datetime
    count: int

    @validator("start_date", pre=True)
    def _date_validate(cls, v: str) -> datetime:
        return datetime.strptime(v, "%Y-%m-%d")


class Achievement(BaseModel):
    id: int
    name: str
    slug: str
    desciption: str
    grouping: str
    icon_url: str
    mode: Gamemode
    ordering: int
    instructions: Optional[str]


class Country(BaseModel):
    code: str
    name: str

    @property
    def flag_emoji(self) -> str:
        r"""Emoji for the flag.

        :return: Unicode emoji representation of the country's flag
        :rtype: str
        """
        return flag_lookup(self.code)


class HTMLBody(BaseModel):
    html: str
    raw: Optional[str]
    bbcode: Optional[str]


class PinAttributes(BaseModel):
    is_pinned: bool
    score_id: int
    score_type: str


class CurrentUserAttributes(BaseModel):
    can_destroy: Optional[bool]
    can_reopen: Optional[bool]
    can_moderate_kudosu: Optional[bool]
    can_resolve: Optional[bool]
    vote_score: Optional[int]
    can_message: Optional[bool]
    can_message_error: Optional[str]
    last_read_id: Optional[int]
    can_new_comment: Optional[bool]
    can_new_comment_reason: Optional[str]
    pin: Optional[PinAttributes]


class CursorModel(BaseModel):
    r"""NOTE: This model is not serializable by orjson directly.

    Use the provided .json() or .dict() methods instead.
    """

    cursor_string: Optional[str]
    next: Optional[partial[Coroutine[Any, Any, CursorModel]]] = Field(exclude=True)
    """Partial function to get the next page of results."""



"""
This module contains models for comment objects.
"""
from __future__ import annotations

from datetime import datetime
from typing import Literal
from typing import Optional

from .base import BaseModel
from .common import CurrentUserAttributes
from .common import CursorModel
from .user import User

__all__ = (
    "Commentable",
    "Comment",
    "CommentBundle",
    "CommentSortType",
)

CommentSortType = Literal["new", "old", "top"]


class Commentable(BaseModel):
    id: Optional[int]
    title: Optional[str]
    url: Optional[str]
    type: Optional[str]
    owner_id: Optional[int]
    owner_title: Optional[str]
    current_user_attributes: Optional[CurrentUserAttributes]


class Comment(BaseModel):
    id: int
    commentable_id: int
    commentable_type: str
    created_at: datetime
    updated_at: datetime
    pinned: bool
    votes_count: int
    replies_count: int
    message: Optional[str]
    message_html: Optional[str]
    deleted_at: Optional[datetime]
    edited_at: Optional[datetime]
    edited_by_id: Optional[int]
    parent_id: Optional[int]
    legacy_name: Optional[str]
    user_id: Optional[int]


class CommentBundle(CursorModel):
    commentable_meta: list[Commentable]
    comments: list[Comment]
    has_more: bool
    included_comments: list[Comment]
    sort: str
    user_follow: bool
    user_votes: list[int]
    users: list[User]
    pinned_comments: Optional[list[Comment]]
    total: Optional[int]
    top_level_count: Optional[int]
    has_more_id: Optional[int]



"""
This module contains models for spotlight objects.
"""
from __future__ import annotations

from datetime import datetime
from typing import Optional

from .base import BaseModel

__all__ = ("Spotlight",)


class Spotlight(BaseModel):
    id: int
    name: str
    mode_specific: bool
    type: str
    start_date: datetime
    end_date: datetime
    participant_count: Optional[int]



"""
This module contains models for replays.
"""
from __future__ import annotations

from datetime import datetime
from enum import IntFlag
from enum import unique
from typing import Any
from typing import Optional

from pydantic import root_validator

from .base import BaseModel
from .gamemode import Gamemode
from .lazer import LazerReplayData
from .mods import Mod
from .mods import Mods
from .score import ScoreStatistics

__all__ = (
    "Replay",
    "ReplayKey",
    "ReplayLifebarEvent",
    "ReplayEvent",
)


def _parse_skip_offset(events: list[ReplayEvent], mods: Mods) -> int:
    """Parse the skip offset from a list of replay events."""
    if len(events) < 2:
        return 0
    if Mod.Autoplay in mods:
        return events[1].time - 100000
    return events[1].time


def _parse_rng_seed(events: list[ReplayEvent]) -> int:
    """Parse the RNG seed from a list of replay events."""
    if len(events) < 2:
        return 0
    return int(events[-1].keys)


@unique
class ReplayKey(IntFlag):
    """Replay key data."""

    K1 = 1 << 0
    K2 = 1 << 1
    K3 = 1 << 2
    K4 = 1 << 3
    K5 = 1 << 4
    K6 = 1 << 5
    K7 = 1 << 6
    K8 = 1 << 7
    K9 = 1 << 8
    K10 = 1 << 9
    K11 = 1 << 10
    K12 = 1 << 11
    K13 = 1 << 12
    K14 = 1 << 13
    K15 = 1 << 14
    K16 = 1 << 15
    K17 = 1 << 16
    K18 = 1 << 17


class ReplayLifebarEvent(BaseModel):
    """Replay lifebar event data."""

    time: int
    hp: float


class ReplayEvent(BaseModel):
    """Replay event data."""

    time: int
    x: float
    y: float
    keys: ReplayKey


class Replay(BaseModel):
    """Replay file data."""

    mode: Gamemode
    version: int
    map_md5: str
    player_name: str
    played_at: datetime
    replay_md5: str
    online_id: int
    score: int
    max_combo: int
    perfect_combo: bool
    mods: Mods
    statistics: ScoreStatistics
    replay_data: list[ReplayEvent]
    lifebar_data: list[ReplayLifebarEvent]
    mod_extras: Optional[float]
    skip_offset: Optional[int]
    rng_seed: Optional[int]
    lazer_replay_data: Optional[LazerReplayData]

    def __repr__(self) -> str:
        return f"<Replay {self.player_name} {self.map_md5}>"

    def __str__(self) -> str:
        return f"{self.player_name} {self.played_at} {self.map_md5} +{self.mods}"

    @root_validator
    def _add_skip_offset(cls, values: dict[str, Any]) -> dict[str, Any]:
        if not values["skip_offset"]:
            values["skip_offset"] = _parse_skip_offset(
                values["replay_data"],
                values["mods"],
            )
        return values

    @root_validator
    def _add_rng_seed(cls, values: dict[str, Any]) -> dict[str, Any]:
        if not values["rng_seed"] and values["version"] >= 2013_03_19:
            values["rng_seed"] = _parse_rng_seed(values["replay_data"])
        return values
